/*=============================================================================
	UnCollision.cpp: Higor's collision octree functions
=============================================================================*/


#include "XC_Engine.h"

#ifndef DISABLE_COLLISION
#include "UnXC_Col.h"

static INT DebugCycles		= 0;
static INT RayCycles		= 0;
static INT TraceCycles		= 0;
static INT TotalCycles		= 0;
static INT GridCycles		= 0;
static INT OctreeGridAssignment[TWO_EXP_OGA];
static FGridData* InternalGridData = NULL;
static FCollisionOctree** CacusNodes;

XC_ENGINE_API FCollisionCacus* CollisionCacus = NULL;


#if 0
	#define check_debug(n) check(n)
	#define guard_debug(n) guard(n)
	#define unguard_debug unguard
	#define unguardf_debug(n) unguardf(n)
#else
	#define check_debug(n) {}
	#define guard_debug(n) {}
	#define unguard_debug {}
	#define unguardf_debug(n) {}
#endif



#define HAS_XC_PRIMITIVE(_a_) _a_->HitActor && _a_->HitActor->IsA(AXC_PrimitiveActor::StaticClass())

#define LIST_BASE_NODES(counter)  		INT counter = 0; \
											for ( FCollisionOctree* Link=Parent ; Link ; Link=Link->Parent ) \
												CacusNodes[counter++] = Link;	\
											CacusNodes[counter++] = this


inline void TwoPointBox( FBox& Box, const FVector4& A, const FVector4& B, const FVector4& Extent);

inline FBox ExpandByExtent( const FBox& Box, const FVector& Extent)
{
	return FBox( Box.Min-Extent, Box.Max+Extent);
}

//We gotta give an index to each piece, num=last in TwoExp
static void CalculateGridIdxTable( INT* IT, INT NT)
{
	INT* Start = IT;
	INT j = 1;
	*IT++ = MAX_OCTREE_DEPTH;
	while ( j<OCTREE_GRID_ARRAY )
	{
		INT add = (IT-Start);
		for ( INT i=0 ; i<add ; i++ )
			*IT++ = MAX_OCTREE_DEPTH - j;
		j++;
	}
	check( IT - Start == NT );
}

inline INT GridIdx( FLOAT Line, const FLOAT& Total)
{
	return OctreeGridAssignment[Clamp( appFloor( Line * (FLOAT)TWO_EXP_OGA * _Reciprocal(Total) ), 0, TWO_EXP_OGA-1)];
}

inline bool GEZero( FLOAT value )
{
	const INT value_as_int = *(INT*)&value;
	return ( value_as_int >= 0 );
}

inline bool DiffSign_NoFPU( FLOAT F1, FLOAT F2)
{
	return ((*(INT*)&F1) ^ (*(INT*)&F2)) < 0;
}

//Nice utilitary function I got from UE sources
//Not as fast as it looks
static UBOOL BoxBoxIntersect(const FBox& box1,const FBox& box2)
{
	if( box1.Min.X > box2.Max.X || box2.Min.X > box1.Max.X )
		return false;
	if( box1.Min.Y > box2.Max.Y || box2.Min.Y > box1.Max.Y )
		return false;
	if( box1.Min.Z > box2.Max.Z || box2.Min.Z > box1.Max.Z )
		return false;
	return true;
}

inline FLOAT Sign( const FLOAT& X)
{
	if ( X > 0 )			return 1.f;
	else if ( X < 0 )		return -1.f;
	return 0.f;
}

inline UBOOL ValidCollidingActor( AActor* Actor)
{
	guard_debug(ValidCollidingActor);
	if ( (UObject::GetIndexedObject( Actor->GetIndex()) != Actor) || (*((INT*)Actor) == 0) )
		return false;
	return Actor->bCollideActors && !Actor->bDeleteMe;
	unguard_debug;
}

inline UBOOL BrushMover( AActor *Actor)
{
	return (Actor->Brush && Actor->bIsMover );
}


//============================================================================//
//============================================================================//
//===                       FCollisionCacus methods                        ===//
//============================================================================//
//============================================================================//

void FCollisionCacus::SetupGrids()
{
	if ( !GridData )
		GridData = new FGridData();
	INT i;
	for ( i=0 ; i<OCTREE_GRID_ARRAY ; i++ )
	{
		DepthGrid[i].Grid = &(GridData->Grid0);
		DepthGrid[i].Grid += GridData->Position[i];
	}
	DepthGrid[0].SetBaseGrid( this);
	for ( i=1 ; i<OCTREE_GRID_ARRAY ; i++ )
		DepthGrid[i].SetGrid( &DepthGrid[i-1]);
	static INT HasGrids = 0;
	if ( !HasGrids++ )
		CalculateGridIdxTable( &OctreeGridAssignment[0], TWO_EXP_OGA);
	
	if ( !InternalGridData )
	{
		InternalGridData = new FGridData();
		CacusNodes = &(InternalGridData->Grid0);
	}
}

FCheckResult* FCollisionCacus::ActorLineCheck( FMemStack& Mem, FVector End, FVector Start, FVector Extent, BYTE ExtraNodeFlags )
{
	guard( FCollisionCacus::ActorLineCheck);
	FCheckResult* Result;
	if ( End == Start )
	{
		Result = ActorPointCheck( Mem, Start, Extent, ExtraNodeFlags);
		FCheckResult* Counted = Result;
		FCheckResult* LastCounted = NULL;
		//Eliminate hits with zero normal
		while ( Counted )
		{
			if ( Counted->Primitive && (Counted->Normal.SizeSquared() < 0.99f) ) //Prevent 'CheckWithActors' crash on movers
			{
				if ( LastCounted )
					LastCounted->Next = Counted->Next;
				else
					Result = Counted->GetNext();
			}
			else
				LastCounted = Counted;
			Counted = Counted->GetNext();
		}
	}
	else
	{
	//	clock( RayCycles);
		FPrecomputedRay Ray( End, Start, Extent); //40 ms every 100 frames (Warzone), this is very expensive
	//	unclock( RayCycles);
		//FAST QUERY! Get nodes from grid based on Start/End
		FCollisionOctree* StartNode = this;
		INT j = GridIdx( Ray.ExtendedBox.Max.X - Ray.ExtendedBox.Min.X, DepthGrid[0].BoxSize.X); //Insane optimization
		INT XYZ[3];
		DepthGrid[j].GetXYZ( Ray.ExtendedBox.Min, XYZ);
		for ( ; j>0 ; j-- )
		{
			FCollisionOctree* TestNode = DepthGrid[j].GetByXYZ( XYZ[0], XYZ[1], XYZ[2]);
			if ( !TestNode || !FitsBox( Ray.ExtendedBox, TestNode->BoundingBox) )
			{
				XYZ[0] = XYZ[0] >> 1;
				XYZ[1] = XYZ[1] >> 1;
				XYZ[2] = XYZ[2] >> 1;
				continue;
			}
			StartNode = TestNode;
			break;
		}
		Result = StartNode->FastLineQuery( Mem, End, Start, Extent, ExtraNodeFlags, Ray, NULL);
	}

	INT Counter = 0;
	FCheckResult* Counted = Result;
	while ( Counted )
	{
		if ( Counter++ > 60 )
		{
			Counted->Next = NULL;
			break;
		}
		Counted = Counted->GetNext();
	}
	return Result;
	unguard;
}

FCheckResult* FCollisionCacus::ActorPointCheck( FMemStack& Mem, FVector Location, FVector Extent, DWORD ExtraNodeFlags )
{
	guard(FCollisionCacus::ActorPointCheck);

	//FAST QUERY! Get nodes from grid based on Location
	FCollisionOctree* StartNode = this;
	INT XYZ[3];
	DepthGrid[MAX_OCTREE_DEPTH].GetXYZ( Location, XYZ);
	UBOOL bZeroExtent = Extent.IsZero();
	for ( INT j=MAX_OCTREE_DEPTH ; j>0 ; j-- )
	{
		FCollisionOctree* TestNode = DepthGrid[j].GetByXYZ( XYZ[0], XYZ[1], XYZ[2]);
		XYZ[0] = XYZ[0] >> 1;
		XYZ[1] = XYZ[1] >> 1;
		XYZ[2] = XYZ[2] >> 1;
		if ( !TestNode ) //Nothing here
			continue;
		if ( !bZeroExtent && !FitsBox( Location, ExpandByExtent(TestNode->BoundingBox, -Extent)) ) //Point not inside reduced box
			continue;
		StartNode = TestNode;
		break;
	}
	return StartNode->FastPointQuery( Mem, Location, Extent, ExtraNodeFlags, NULL);
	unguard;
}

FCheckResult* FCollisionCacus::ActorRadiusCheck( FMemStack& Mem, FVector Location, FLOAT Radius, DWORD ExtraNodeFlags )
{
	guard( FCollisionCacus::ActorRadiusCheck)
	return (FCheckResult*) RadiusQuery( Mem, Location, Radius, ExtraNodeFlags, NULL);
	unguard;
}

FCheckResult* FCollisionCacus::ActorEncroachmentCheck( FMemStack& Mem, AActor* Actor, FVector Location, FRotator Rotation, DWORD ExtraNodeFlags )
{
	guard( FCollisionCacus::ActorEncroachmentCheck);
	if ( BrushMover(Actor) )
	{
		Exchange( Actor->Location, Location);
		FBox Box = Actor->Brush->GetCollisionBoundingBox( Actor);

		FCheckResult* Result = MEncroachQuery( Mem, Actor, Box);
		Exchange( Actor->Location, Location);
		return Result;
	}
	//Added a bCollideActors check here, LCWeapons' position advancer was crashing due to not having this check...
	else if ( Actor->bCollideActors && (Actor->bBlockPlayers || Actor->bBlockActors) ) //Griding did nothing!!! let's stop the CPU usage here
	{
		const FBox ActorBox( Location-Actor->GetCylinderExtent(), Location+Actor->GetCylinderExtent() );

		FCollisionOctree* StartNode = this;
		INT j = GridIdx( ActorBox.Max.X - ActorBox.Min.X, DepthGrid[0].BoxSize.X); //Insane optimization
		INT XYZ[3];
		DepthGrid[j].GetXYZ( Location, XYZ);
		for ( ; j>0 ; j-- )
		{
			FCollisionOctree* TestNode = DepthGrid[j].GetByXYZ( XYZ[0], XYZ[1], XYZ[2]);
			if ( !TestNode || !FitsBox( ActorBox, TestNode->BoundingBox) )
			{
				XYZ[0] = XYZ[0] >> 1;
				XYZ[1] = XYZ[1] >> 1;
				XYZ[2] = XYZ[2] >> 1;
				continue;
			}
			StartNode = TestNode;
			break;
		}
		return StartNode->FastAEncroachQuery( Mem, Actor, Location, ActorBox);
	}
	return NULL;
	unguard;
}



//Quick Add actor/mover
//Also, don't make children nodes if there's 3 or less actors in box
void FCollisionCacus::AddActor( AActor *Actor )
{
	guard( FCollisionCacus::AddActor);
	check_debug( Actor);
	static UBOOL bNextLog = 0;
	if ( !ValidCollidingActor(Actor) )
		return;
	if ( appIsNan(Actor->Location.X) ) //Normally if one of the components is bad, the others are bad
		return;
	FCacusActorInfo NewInfo( Actor); //Construct info for this actor
	FVector PrimSize = NewInfo.Box.Max - NewInfo.Box.Min;
	FCollisionOctree* TestNode = this; //Default node

	//Use the grid no matter what
	//New procedural loops
	guard(NodeLoops);
	//Discard grids with insufficiently sized boxes
//	INT j=MAX_OCTREE_DEPTH;
	INT j;
	INT XYZ[3];
	for ( j=GridIdx( PrimSize.X, DepthGrid[0].BoxSize.X) ; j>0 ; j-- ) //X fit check taken care of here
		if ( (DepthGrid[j].BoxSize.Y >= PrimSize.Y) && (DepthGrid[j].BoxSize.Z >= PrimSize.Z) ) //Y,Z fit check taken care of here
		{
			DepthGrid[j].GetXYZ( NewInfo.Location, XYZ);
			break;
		}

	//See if specific box in this grid contains the primitive
	while ( j > 0 )
	{
		FBox SlotBox = DepthGrid[j].GetBox( XYZ[0], XYZ[1], XYZ[2]);
		if ( FitsBox( NewInfo.Box, SlotBox) )
			break;
		XYZ[0] = XYZ[0] >> 1;
		XYZ[1] = XYZ[1] >> 1;
		XYZ[2] = XYZ[2] >> 1;
		j--;
	}

	NewInfo.TopDepth = j; //This is as far as this can go
	
	//Find lowest depth existing node (may be the one we're looking for)
	if ( j > 0 )
	{
		TestNode = DepthGrid[j].GetByXYZ( XYZ[0], XYZ[1], XYZ[2]);
		if ( !TestNode )
		{
			INT k = j;
			//Find lowest depth existing node, master guaranteed to be found
			do
			{
				INT Diff = j - (--k);
				TestNode = DepthGrid[k].GetByXYZ( XYZ[0] >> Diff, XYZ[1] >> Diff, XYZ[2] >> Diff);
			}
			while ( !TestNode );
			check_debug(TestNode);
			//Create ONE subnode if we have enough actors
			if ( TestNode->Actors.Num() + TestNode->BrushMovers.Num() >= MIN_ACTORS_FOR_SUBNODE )
			{
				INT Diff = j - (++k);
				XYZ[0] = XYZ[0] >> Diff;
				XYZ[1] = XYZ[1] >> Diff;
				XYZ[2] = XYZ[2] >> Diff;
				TestNode = new FCollisionOctree( TestNode, XYZ);
			}
		}
	}
	if ( BrushMover( Actor) )
		TestNode->BrushMovers.AddItem( NewInfo);
	else
		TestNode->Actors.AddItem( NewInfo);
	Actor->CollisionTag = TestNode->MIndex;
	Actor->ColLocation = Actor->Location;
	unguardf( (TEXT("%s"), Actor->GetName() ) );
	unguard;
}


//Quick remove actor/mover
//No sanity checks performed for speed reasons
//Does not perform memory deallocations
void FCollisionCacus::RemoveActor( AActor *Actor )
{
	guard( FCollisionCacus::RemoveActor);

	if ( Actor->CollisionTag == -1 ) //Double removal issue, disregard
		return;
	
	//Anomaly:
	// Unhashed actor being removed without being added
	FCollisionOctree* Node = GridData->FullList( Actor->CollisionTag);
	if ( !Node )
	{
		debugf( NAME_XC_Engine, TEXT("Octree anomaly #1 [NO_NODE] with actor: %s, %i"), Actor->GetName(), Actor->CollisionTag );
		return;
	}

	INT i=0;
	TArray<FCacusActorInfo>* AArray = ((TArray<FCacusActorInfo>*) &(Node->Actors)) + BrushMover( Actor);
	INT* ArrayNum = ((INT*) AArray) + 1;
	while ( i<*ArrayNum )
	{
		if ( (*AArray)(i).Actor == Actor )
		{
			(*ArrayNum)--;
			if ( i != (*ArrayNum) )
				(*AArray)(i) = (*AArray)(*ArrayNum);
			Actor->CollisionTag = -1;
			return;
		}
		i++;
	}//Note, node 0 always checked on actors with CollisionTag=0, so if it passes, it's a fail

	//Bug!
	// Actor is supposed to be in this node!!
	if ( Actor->CollisionTag == 0 ) //EXPECTED
	{
		debugf( NAME_XC_Engine, TEXT("Octree anomaly #2 [UNHASHED] with actor: %s"), Actor->GetName() );
		return;
	}
	
	FCollisionOctree* RealNode = SlowRemoveActor(Actor);
	if ( RealNode ) //THIS IS VERY BAD
	{
		debugf( NAME_XC_Engine, TEXT("Octree anomaly #3 [INCORRECT_ALLOCATION] with actor: %s (%i -> %i) | %i"), Actor->GetName(), Node->MIndex, RealNode->MIndex, BrushMover(Actor) );
		if ( Node->MIndex == RealNode->MIndex && Node != RealNode )
			debugf( NAME_XC_Engine, TEXT("Octree anomaly #5 [NODE_REDUNDANCY]: two nodes spawned with same MIndex!!!!"));
		return;
	}
	
	debugf( NAME_XC_Engine, TEXT("Octree anomaly #4 [UNHASHED_CRITICAL] with actor: %s"), Actor->GetName() );
	unguard;
}

//Does nothing in default hash
void FCollisionCacus::CheckActorNotReferenced( AActor* Actor )
{
}

void FCollisionCacus::Tick()
{
	Cleanup();
	if ( DebugCycles++ > 100 )
	{
//		debugf( NAME_Log, TEXT("Trace: T=%03.1f, R=%03.1f, L=%03.1f"), GSecondsPerCycle*1000 * TotalCycles, GSecondsPerCycle*1000 * RayCycles, GSecondsPerCycle*1000 * TraceCycles);
//		debugf( NAME_Log, TEXT("Cycles measured: %03.1f"), GSecondsPerCycle*1000 * RayCycles);
		RayCycles = 0;
		TraceCycles = 0;
		TotalCycles = 0;
		DebugCycles = 0;
//		GridCycles = 0;
	}
}

void FCollisionCacus::Exit()
{
	guard( FCollisionCacus::Exit );
	FCollisionOctree::Exit();
	for ( INT i=0 ; i<OCTREE_GRID_ARRAY ; i++ )
		DepthGrid[i].Grid = NULL;

	if (GridData)
		delete GridData;
	GridData = NULL;

	if (InternalGridData)
		delete InternalGridData;
	InternalGridData = NULL;

	unguard;
}


FCollisionCacus::~FCollisionCacus()
{
	guard_debug( ~FCollisionCacus );
	Exit();
	CollisionCacus = NULL;
	unguard_debug;
}

//============================================================================//
//============================================================================//
//===                       FCollisionOctree methods                       ===//
//============================================================================//
//============================================================================//


FCollisionOctree::FCollisionOctree( FCollisionOctree* InParent, INT* XYZ)
	:	Master(InParent->Master)
	,	Parent(InParent)
//	,	BoundingBox(InParent->BoundingBox)
	,	CurDepth(InParent->CurDepth + 1)
	,	CleanupTimer(20)
	,	ChildCount(0)
	,	XPos( XYZ[0] )
	,	YPos( XYZ[1] )
	,	ZPos( XYZ[2] )
	,	TotalChildren(0)
	,	Actors(0)
	,	BrushMovers(0)
{
	check_debug( InParent->ChildCount < 8 );
	check_debug( !InParent->ReducedChildren[InParent->ChildCount] );

	appMemzero( &ReducedChildren[0], 4*8 ); //Child array
	Parent->ReducedChildren[InParent->ChildCount++] = this;
	for ( FCollisionOctree* Link=Parent ; Link ; Link=Link->Parent )
		Link->TotalChildren++;

	//Construct box using grid
	BoundingBox = Master->DepthGrid[CurDepth].GetBox( XPos, YPos, ZPos);
/*	if ( !FitsBox( BoundingBox, Parent->BoundingBox) )
	{
		debugf( NAME_Log, TEXT("ERROR: Node created outside of parent"));
		debugf( NAME_Log, TEXT("L_BOX(%i): (%f,%f,%f)(%f,%f,%f)"), CurDepth, BoundingBox.Min.X, BoundingBox.Min.Y, BoundingBox.Min.Z, BoundingBox.Max.X, BoundingBox.Max.Y, BoundingBox.Max.Z);
		debugf( NAME_Log, TEXT("H_BOX(%i): (%f,%f,%f)(%f,%f,%f)"), Parent->CurDepth, Parent->BoundingBox.Min.X, Parent->BoundingBox.Min.Y, Parent->BoundingBox.Min.Z, Parent->BoundingBox.Max.X, Parent->BoundingBox.Max.Y, Parent->BoundingBox.Max.Z);
	}*/
	Master->DepthGrid[CurDepth].Register( this);

	//Grab actors from parent node if possible
	guard(GrabFromParent);
	INT i;
	for ( i=0 ; i<Parent->Actors.Num() ; i++ )
		if ( (Parent->Actors(i).TopDepth >= CurDepth) && FitsBox( Parent->Actors(i).Box, BoundingBox) )
		{
			Actors.AddItem( Parent->Actors(i) );
			Parent->Actors(i).Actor->CollisionTag = MIndex;
			Parent->Actors.Remove( i--, 1);
		}
	for ( i=0 ; i<Parent->BrushMovers.Num() ; i++ )
		if ( (Parent->BrushMovers(i).TopDepth >= CurDepth) && FitsBox( Parent->BrushMovers(i).Box, BoundingBox) )
		{
			BrushMovers.AddItem( Parent->BrushMovers(i) );
			Parent->BrushMovers(i).Actor->CollisionTag = MIndex;
			Parent->BrushMovers.Remove( i--, 1);
		}
	unguard;
}


//XC_GameEngine interface
FCacusActorInfo* FCollisionOctree::FindActor( FName ByName)
{
	guard( FCollisionOctree::FindActor);
	INT i;
	for ( i=0 ; i<Actors.Num() ; i++ )
	{
		if ( Actors(i).Actor->GetFName() == ByName )
			return &Actors(i);
	}
	for ( i=0 ; i<BrushMovers.Num() ; i++ )
	{
		if ( BrushMovers(i).Actor->GetFName() == ByName )
			return &BrushMovers(i);
	}
	FCacusActorInfo* Result = NULL;
	for ( i=0 ; i<ChildCount ; i++ )
	{
		Result = ReducedChildren[i]->FindActor( ByName);
		if ( Result )
			break;
	}
	return Result;
	unguard;
}

FCollisionOctree* FCollisionOctree::SlowRemoveActor( AActor* Other)
{
	guard(FCollisionOctree::SlowRemoveActor);
	INT i=0;
	for ( ; i<Actors.Num() ; i++ )
		if ( Actors(i).Actor == Other )
		{
			Actors.Remove( i);
			Other->CollisionTag = -1;
			return this;
		}
	for ( i=0 ; i<ChildCount ; i++ )
	{
		FCollisionOctree* Node = ReducedChildren[i]->SlowRemoveActor( Other);
		if ( Node )
			return Node;
	}
	return NULL;
	unguard;
}

void FCollisionOctree::Cleanup()
{
	guard( FCollisionOctree::Cleanup);
	INT i=0;
	//Cleanup child nodes first
	guard( CleanupChild);
	for ( i=0 ; i<ChildCount ; i++ )
	{
		check( ReducedChildren[i] );
		check( ReducedChildren[i]->Parent == this);
		ReducedChildren[i]->Cleanup();
		if ( ReducedChildren[i]->ChildCount || ReducedChildren[i]->Actors.Num() || ReducedChildren[i]->BrushMovers.Num() )
			continue;
		if ( --(ReducedChildren[i]->CleanupTimer) <= 0 )
		{
			ReducedChildren[i]->Exit();
			delete ReducedChildren[i];
			ReducedChildren[i] = ReducedChildren[--ChildCount];
			ReducedChildren[ChildCount] = NULL;
		}
	}
	unguardf( (TEXT("Depth=%i, i=%i, Idx=%i"), CurDepth, i, ReducedChildren[i]->ChildIndex));

	//Move my actors to parent node if we're holding too few actors, parent's Cleanup will delete me later
	//BTW, on this new format... TopDepth doesn't need to be edited!!
	guard( ParentCL);
	if ( !ChildCount && CurDepth )
	{
		if ( (Actors.Num() + BrushMovers.Num() + Parent->Actors.Num() + Parent->BrushMovers.Num()) < MIN_ACTORS_FOR_SUBNODE )
		{
			if ( --CleanupTimer <= 1 )
			{
				INT j = 0;
				if ( Actors.Num() )
				{
					j = Parent->Actors.Num();
					Parent->Actors.Add( Actors.Num() );
					appMemcpy( &Parent->Actors(j), &Actors(0), sizeof(struct FCacusActorInfo) * Actors.Num() );
					for ( i=0 ; i<Actors.Num() ; i++ )
						Parent->Actors(i+j).Actor->CollisionTag = Parent->MIndex;
					Actors.Empty();
				}
				if ( BrushMovers.Num() )
				{
					j = Parent->BrushMovers.Num();
					Parent->BrushMovers.Add( BrushMovers.Num() );
					appMemcpy( &Parent->BrushMovers(j), &BrushMovers(0), sizeof(struct FCacusActorInfo) * BrushMovers.Num() );
					for ( i=0 ; i<BrushMovers.Num() ; i++ )
						Parent->BrushMovers(i+j).Actor->CollisionTag = Parent->MIndex;
					BrushMovers.Empty();
				}
			}
		}
		return;
	}
	unguard;

	//Cleanup no longer moves actors into deeper nodes
	guard( CleanupActors);
	i=0;
	while ( i<Actors.Num() )
	{
		if ( !ValidCollidingActor( Actors(i).Actor) )
		{
			Actors.Remove( i);
			continue;
		}
		i++;
	}
	i=0;
	while ( i<BrushMovers.Num() )
	{
		if ( !ValidCollidingActor( BrushMovers(i).Actor) )
		{
			BrushMovers.Remove( i);
			continue;
		}
		i++;
	}
	unguard;

	unguard;
}

//Recursive actor iteration
//Obtain actors in a specific radius
FIteratorActorList* FCollisionOctree::RadiusQuery( FMemStack& Mem, const FVector& Location, const FLOAT& Radius, const DWORD& ExtraNodeFlags, FIteratorActorList* PrevLink )
{
	guard( FCollisionOctree::RadiusQuery);
	if ( CurDepth && !FitsBox( Location, BoundingBox.ExpandBy(Radius)) )
		return PrevLink;
	FLOAT RSq = Radius * Radius;
	INT i;
	for ( i=0 ; i<Actors.Num() ; i++ )
		if ( ValidCollidingActor(Actors(i).Actor) && (FDistSquared( Actors(i).Location, Location) <= RSq) )
			PrevLink = new(Mem)FIteratorActorList( PrevLink, Actors(i).Actor);
	for ( i=0 ; i<BrushMovers.Num() ; i++ )
		if ( ValidCollidingActor(BrushMovers(i).Actor) && (FDistSquared( BrushMovers(i).Location, Location) <= RSq) )
			PrevLink = new(Mem)FIteratorActorList( PrevLink, BrushMovers(i).Actor);
	for ( i=0 ; i<ChildCount ; i++ )
		PrevLink = ReducedChildren[i]->RadiusQuery( Mem, Location, Radius, ExtraNodeFlags, PrevLink);
	return PrevLink;
	unguardf( (TEXT("Depth %i"), CurDepth) );
}

/** v16
 * Non recursive version of PointQuery
 * Single alloc, linear and more CPU cache friendly
 * Also precaches point box to allow excessive FPU usage
 */
FCheckResult* FCollisionOctree::FastPointQuery( FMemStack& Mem, const FVector& Location, const FVector& Extent, const DWORD& ExtraNodeFlags, FCheckResult* PrevLink)
{
	guard( FCollisionOctree::FastPointQuery);
	
	//Cache an extent box to avoid expanding every single actor's box
	const FBox ExtentBox( Location - Extent, Location + Extent);
	
	//Add parent/self nodes
	LIST_BASE_NODES(CurRef);
	for ( INT ChildStart=CurRef-1 ; ChildStart < CurRef ; ChildStart++ )
	{
		FCollisionOctree* CurNode = CacusNodes[ChildStart];
		for ( INT i=0 ; i<CurNode->ChildCount ; i++ ) //Register new child nodes into the list
			if ( BoxBoxIntersect( ExtentBox, CurNode->ReducedChildren[i]->BoundingBox) )
				CacusNodes[CurRef++] = CurNode->ReducedChildren[i];
	}
	INT MaxNodes = CurRef;
	
	INT i;
	guard( NodeLoop);
	for ( CurRef=0 ; CurRef<MaxNodes ; CurRef++ )
	{
		FCollisionOctree* Link = CacusNodes[CurRef];
		guard_debug(ActorCheck);
		for ( i=0 ; i<Link->Actors.Num() ; i++ )
		{
			if ( !ValidCollidingActor(Link->Actors(i).Actor) )
				continue;
			if ( !BoxBoxIntersect( ExtentBox, Link->Actors(i).Box) ) //Quick box rejects, we don't need to perform Z check later
				continue;
			if ( Square( Link->Actors(i).Location.X - Location.X) + Square( Link->Actors(i).Location.Y - Location.Y) <= Square( Link->Actors(i).Actor->CollisionRadius + Extent.X) )
			{
				UBOOL bNeedsNormal = 1;
				FCheckResult Hit(1.f, PrevLink);
				Hit.Location = Link->Actors(i).Location; //New hitlocation, gives opportunity to adjust using primitive
				Hit.Actor = Link->Actors(i).Actor;
				if ( HAS_XC_PRIMITIVE(Hit.Actor) )
				{
					AXC_PrimitiveActor* Primitive = (AXC_PrimitiveActor*)Hit.Actor->HitActor;
					Primitive->HitActor = Hit.Actor;
					if ( !(Primitive->bNoPointCheck || (Primitive->bNoExtentCheck && !Extent.IsZero())) )
					{	if ( !Primitive->InContact( Mem, Location, Extent, &Hit) )
							continue;
						bNeedsNormal = 0;
					}
					else if ( !Primitive->bFallBackToCylinder )
						continue;
				}
				if ( bNeedsNormal )
				{
					guard_debug(MakeNormal);
					Hit.Normal = (Hit.Location - Location).SafeNormal();
					if ( Hit.Normal.Z > 0.5 )
						Hit.Location.Z += Extent.Z;
					else if ( Hit.Location.Z < -0.5 )
						Hit.Location.Z -= Extent.Z;
					else
						Hit.Location += (Hit.Normal) * Extent.X;
					unguard_debug;
				}
				PrevLink = new(Mem)FCheckResult( 1.f, PrevLink);
				appMemcpy( PrevLink, &Hit, sizeof(Hit));
			}
		}
		unguard_debug;
		guard_debug(MoverCheck);
		for ( i=0 ; i<Link->BrushMovers.Num() ; i++ )
		{
			if ( Link->BrushMovers(i).Actor->Brush && BoxBoxIntersect( ExtentBox, Link->BrushMovers(i).Box) ) //Better box rejects
			{
				FCheckResult Hit(1.f, PrevLink);
				if ( !Link->BrushMovers(i).Actor->Brush->PointCheck( Hit, Link->BrushMovers(i).Actor, Location, Extent, ExtraNodeFlags) )
				{
					PrevLink = new(Mem)FCheckResult(1.f,PrevLink);
					appMemcpy( PrevLink, &Hit, sizeof( Hit) );
					PrevLink->Time -= 0.0001f;
					PrevLink->Location += PrevLink->Normal * 2.02f; //Prevent 'CheckForActors' crash (?)
				}
			}
		}
		unguard_debug;
	}
	unguardf( (TEXT("Node %i/%i"),CurRef,MaxNodes) );
	return PrevLink;
	unguardf( (TEXT("At: (%f,%f,%f)"), Location.X, Location.Y, Location.Z) );
}

//Recursive actor iteration
//Obtains actors touching a brush
FCheckResult* FCollisionOctree::MEncroachQuery( FMemStack& Mem, AActor* Actor, const FBox& MoverBox, FCheckResult* PrevLink)
{
	guard_debug( Deep);
	INT i;
	for ( i=0 ; i<Actors.Num() ; i++ )
	{	//this bCollideWorld reject is responsible for a massive speedup!
		if ( !Actors(i).Actor->bCollideWorld || !Actors(i).Actor->bBlockActors || !BoxBoxIntersect( MoverBox, Actors(i).Box) ) //Quick box rejects
			continue;
		AActor* TestActor = Actors(i).Actor;
		if ( !ValidCollidingActor(TestActor) ) //I assume Cleanup handles this
			continue;
		FCheckResult TestHit(1.0);
		if ( Actor->Brush->PointCheck( TestHit, Actor, Actors(i).Location, TestActor->GetCylinderExtent(), 0 )==0 )
		{
			TestHit.Actor = TestActor;
			TestHit.Primitive = NULL;
			TestHit.Next = PrevLink;
			PrevLink = new(Mem)FCheckResult;
			*PrevLink = TestHit;
		}
	}
	for ( i=0 ; i<ChildCount ; i++ )
		if ( BoxBoxIntersect( MoverBox, ReducedChildren[i]->BoundingBox) )
			PrevLink = ReducedChildren[i]->MEncroachQuery( Mem, Actor, MoverBox, PrevLink);
	return PrevLink;
	unguardf_debug( (TEXT("Depth %i"), CurDepth) );
}

/** v16
 * Non recursive version of AEncroachQuery
 * Single alloc, linear and more CPU cache friendly
 */

FCheckResult* FCollisionOctree::FastAEncroachQuery( FMemStack& Mem, AActor* Actor, const FVector& Location, const FBox& ActorBox, FCheckResult* PrevLink)
{
	guard(FCollisionOctree::FastAEncroachQuery);
	//Add parent/self nodes
	LIST_BASE_NODES(CurRef);
	for ( INT ChildStart=CurRef-1 ; ChildStart < CurRef ; ChildStart++ )
	{
		FCollisionOctree* CurNode = CacusNodes[ChildStart];
		for ( INT i=0 ; i<CurNode->ChildCount ; i++ ) //Register new child nodes into the list
			if ( BoxBoxIntersect( ActorBox, CurNode->ReducedChildren[i]->BoundingBox) )
				CacusNodes[CurRef++] = CurNode->ReducedChildren[i];
	}
	INT MaxNodes = CurRef;

	for ( CurRef=0 ; CurRef<MaxNodes ; CurRef++ )
	{
		guard(NodeLoop);
		FCollisionOctree* Link = CacusNodes[CurRef];
		for ( INT i=0 ; i<Link->Actors.Num() ; i++ )
		{
			guard(ActorLoop);
			if ( !BoxBoxIntersect( ActorBox, Link->Actors(i).Box) ) //Quick box rejects, we don't need to perform Z check later
				continue;
			AActor* TestActor = Link->Actors(i).Actor;
			if ( TestActor != Actor && (FVector(Location.X-TestActor->Location.X,Location.Y-TestActor->Location.Y,0.f).SizeSquared2D() < Square( Actor->CollisionRadius+TestActor->CollisionRadius))  )
			{
				if ( HAS_XC_PRIMITIVE(TestActor) )
				{
					AXC_PrimitiveActor* Primitive = (AXC_PrimitiveActor*)TestActor->HitActor;
					Primitive->HitActor = TestActor;
					if ( !Primitive->bNoEncroachmentCheck && !Primitive->bNoExtentCheck )
					{	if ( !Primitive->InContact( Mem, Location, Actor->GetCylinderExtent(), NULL) )
							continue;
					}
					else if ( !Primitive->bFallBackToCylinder )
						continue;
				}
				PrevLink = new(Mem)FCheckResult( 0.f, PrevLink);
				PrevLink->Actor = TestActor;
				PrevLink->Location = TestActor->Location;
				PrevLink->Primitive = NULL;
			}
			unguardf( (TEXT("Actor %i/%i"),i,Link->Actors.Num() ) );
		}
		unguardf( (TEXT("Node %i/%i"),CurRef,MaxNodes) );
	}

	return PrevLink;
	unguardf( (TEXT("%s at (%f,%f,%f)"), Actor->GetFullName(), Actor->Location.X, Actor->Location.Y, Actor->Location.Z));
}


/** v16
 * Non recursive version of LineQuery
 * Single alloc, linear and more CPU cache friendly
 * Obtains actors touching a line
 * TODO: Implement directly in ActorLineCheck!!!!
 */
FCheckResult* FCollisionOctree::FastLineQuery
(
 FMemStack& Mem
, const FVector& End
, const FVector& Start
, const FVector& Extent
, const BYTE& ExtraNodeFlags
, FPrecomputedRay& Ray
,  FCheckResult* PrevLink
)
{
	guard(FCollisionOctree::FastLineQuery);

	//Add parent/self nodes
	LIST_BASE_NODES(CurRef);

	guard_debug(ChildLoop)
	//Add child nodes
	for ( INT ChildStart=CurRef-1 ; ChildStart < CurRef ; ChildStart++ )
	{
		FCollisionOctree* CurNode = CacusNodes[ChildStart];
		for ( INT i=0 ; i<CurNode->ChildCount ; i++ ) //Register new child nodes into the list
		{
			check_debug(CurNode->ReducedChildren[i]);
			if ( Ray.IntersectsBox(CurNode->ReducedChildren[i]->BoundingBox) )
				CacusNodes[CurRef++] = CurNode->ReducedChildren[i];
		}
	}
	unguardf_debug( (TEXT("CurRef=%i, Depth=%i, TotalChildren=%i"), CurRef, CurDepth, TotalChildren) );
	INT MaxNodes = CurRef;
	
	//Perform the line checks
	for ( CurRef=0 ; CurRef<MaxNodes ; CurRef++ )
	{
		INT ActorCount = CacusNodes[CurRef]->Actors.Num();
		if ( ActorCount )
		{
			FVector4 HitLocation; //TODO: use FVector4!!!
			FVector4 HitNormal;
			AXC_PrimitiveActor* Primitive;
			guard_debug(ActorLoop);
			for ( INT i=0 ; i<ActorCount ; i++ )
			{
				AActor* Actor = CacusNodes[CurRef]->Actors(i).Actor;
				if ( !ValidCollidingActor(Actor) )
					continue;
				Primitive = NULL;
				if ( !Ray.HitsGCylActor( Actor, &HitLocation, &HitNormal, Primitive) )
					continue;
				guard_debug( PrimitiveAssess);
				if ( Primitive )
				{	Primitive->HitActor = Actor;
					if ( !(Primitive->bNoLineCheck || (Primitive->bNoExtentCheck && !Ray.Extent.IsZero())) )
					{	if ( !Primitive->RayTrace(Mem,Ray,&HitLocation,&HitNormal) )
							continue;
					}
					else if ( !Primitive->bFallBackToCylinder )
						continue;
				}
				unguard_debug;
				guard_debug( LinkGeneration);
				FLOAT Time = LengthUsingDirA( &Ray.Org, &HitLocation, &Ray.Dir);
				if ( Time < 0.f )
					Time = 0.f;
				if ( Time < Ray.Dist ) //NaN check ruled out, Ray.Dist is positive and a zero dist ray always fails
				{
	//				PrevLink = new(Mem) FCheckResult( Clamp( ((HitLocation-Ray.Org) | Ray.Dir) * _Reciprocal(Ray.Dist), 0.f, 1.f) , PrevLink);
					PrevLink = new(Mem) FCheckResult( Time * _Reciprocal(Ray.Dist), PrevLink);
					PrevLink->Actor = Actor;
					PrevLink->Location = HitLocation;
					PrevLink->Normal = HitNormal;
				}
				unguardf_debug( (TEXT("%s, hit: (%f,%f,%f)"), Actor->GetName(), HitLocation.X, HitLocation.Y, HitLocation.Z));
			}
			unguard_debug;
		}
		ActorCount = CacusNodes[CurRef]->BrushMovers.Num();
		for ( INT i=0 ; i<ActorCount ; i++ )
			if ( CacusNodes[CurRef]->BrushMovers(i).Actor->Brush && Ray.IntersectsBox( CacusNodes[CurRef]->BrushMovers(i).Box) )
			{
				FCheckResult Hit(0);
				if ( !CacusNodes[CurRef]->BrushMovers(i).Actor->Brush->LineCheck( Hit, CacusNodes[CurRef]->BrushMovers(i).Actor, End, Start, Extent, ExtraNodeFlags) )
				{
					FCheckResult* NewLink = new(Mem)FCheckResult(Hit);
					NewLink->GetNext() = PrevLink;
					PrevLink = NewLink;
					NewLink->Time -= 0.0001f;
					NewLink->Location += NewLink->Normal * 2.02f; //Prevent 'CheckForActors' crash (?)
				}
			}
	}
	return PrevLink;
	unguard;
}


//Manual cleanup, needed in v451
void FCollisionOctree::Exit()
{
	guard( FCollisionOctree::Exit );
	SafeEmpty( Actors);
	SafeEmpty( BrushMovers);
	unguard;
}

FCollisionOctree::~FCollisionOctree()
{
	guard( Destructor);

	INT i;
	FCollisionOctree::Exit();
	if ( Parent )
	{
		//This requires HAX deinit
		for ( FCollisionOctree* Link=Parent ; Link ; Link=Link->Parent )
			Link->TotalChildren -= ChildCount+1;
	}

	//HAX deinit, useful for fast mapswitch
	for ( i=0 ; i<ChildCount ; i++ )
	{
		ReducedChildren[i]->Parent = NULL;
		delete ReducedChildren[i];
	}

	if ( Master && CurDepth )
		Master->DepthGrid[CurDepth].Remove( this); //No need to clear by MIndex!

	unguard;
}


//============================================================================//
//============================================================================//
//===                       FPrecomputedRay methods                        ===//
//============================================================================//
//============================================================================//

FPrecomputedRay::FPrecomputedRay( const FVector& A, const FVector& B, const FVector& TExtent)
	:	End( A, 1.f)
	,	Org( B, 1.f)
//	,	Dir( A-B, 1.f)
	,	Extent( TExtent)
{
	guard(FPrecomputedRay::FPrecomputedRay);
	TwoPointBox( ExtendedBox, End, Org, Extent);
	FVector4 Segment( A-B, 0.f); //Ugly but necessary
#if ASMLINUX
	//gcc can't inline either UnsafeNormal3 or Dot4... gotta figure out what's upper_bound
	Dir = Segment.UnsafeNormal();
	Dir.W = -1.f;
	Dist = Segment | Dir;
#else
	Dir = Segment.UnsafeNormal3( -1.f);
	Dist = Segment.Dot4( Dir);
#endif
	
	//Don't use Inv on zero components
	if ( Square(Segment.X) + Square(Segment.Y) > SMALL_NUMBER ) //Not the best but it works
	{
		Inv.X = Dir.X == 0.f ? 0.f : _Reciprocal( Dir.X);
		Inv.Y = Dir.Y == 0.f ? 0.f : _Reciprocal( Dir.Y);
		CollisionCacus->PrecomputedCoords.Origin = Org;
		CollisionCacus->PrecomputedCoords.XAxis = _UnsafeNormal2D( Dir.X, Dir.Y);
		CollisionCacus->PrecomputedCoords.YAxis = FVector( -CollisionCacus->PrecomputedCoords.XAxis.Y, CollisionCacus->PrecomputedCoords.XAxis.X, 0.f); //Rotate 90 to right in 2D
		if ( Abs(Dir.Z) <= SMALL_NUMBER )
		{
			Inv.Z = 0.f;
			TraceFunc = &FPrecomputedRay::HitsHCylActor;
		}
		else
		{
			Inv.Z = _Reciprocal( Dir.Z);
			TraceFunc = &FPrecomputedRay::HitsCylActor;
		}
	}
	else
	{
		Inv = Dir; //Because 1/1 = 1
		TraceFunc = &FPrecomputedRay::HitsVCylActor;
	}
	unguard;
}


FORCEINLINE UBOOL FPrecomputedRay::HitsGCylActor( AActor* Actor, FVector4* HitLocation, FVector4* HitNormal, AXC_PrimitiveActor* & Primitive)
{
/*		clock( TraceCycles);
		UBOOL Result = (*this.*TraceFunc)( Actor, HitLocation, HitNormal);
		unclock( TraceCycles);
		return Result;*/
	return (*this.*TraceFunc)( Actor, HitLocation, HitNormal, Primitive);
}

UBOOL FPrecomputedRay::HitsVCylActor( AActor* Actor, FVector4* HitLocation, FVector4* HitNormal, AXC_PrimitiveActor* & Primitive)
{
	guard_debug(FPrecomputedRay::HitsVCylActor);
	if ( Square(Actor->Location.X - Org.X) + Square(Actor->Location.Y - Org.Y) > Square(Extent.X + Actor->CollisionRadius) )
		return 0; //Out of cylinder radius
	FLOAT AdjZ = Actor->Location.Z - Org.Z;
	FLOAT DiffZ = End.Z - Org.Z;
	if ( HAS_XC_PRIMITIVE(Actor) ) //Grab primitive before backtrace check
		Primitive = (AXC_PrimitiveActor*) Actor->HitActor;
	FLOAT TotalHeight = Actor->CollisionHeight + Extent.Z;
	if ( Abs(AdjZ) < TotalHeight ) //Start inside cylinder?
	{
		if ( Primitive && Primitive->bBackTrace )
		{
			*HitLocation = Org;
			return 1;
		}
		return 0; //Start point already colliding
	}
//	if ( fast_sign_nozero(AdjZ) != fast_sign_nozero(DiffZ) )
	if ( (*(INT*)&AdjZ ^ *(INT*)&DiffZ) < 0) //Diff sign check, no FPU stack usage
		return 0; //Opposite directions
	if ( Abs(AdjZ) - TotalHeight > Abs(DiffZ) )
		return 0; //No reach
	*HitNormal = FVector(0,0,-fast_sign_nozero(DiffZ));
	*HitLocation = Org;
	HitLocation->Z = Actor->Location.Z - TotalHeight * fast_sign_nozero(AdjZ);
	return 1;
	unguardf_debug( (Actor->GetName()) );
}

//This is most commonly used during PHYS_Walking movement
//(Dist = HLength) so i don't need to measure trace length again!!!
UBOOL FPrecomputedRay::HitsHCylActor( AActor* Actor, FVector4* HitLocation, FVector4* HitNormal, AXC_PrimitiveActor* & Primitive)
{
	guard_debug(FPrecomputedRay::HitsHCylActor);
	//Z reject (extent smaller-than difference)
	if ( Square(Extent.Z + Actor->CollisionHeight) < Square( Actor->Location.Z - Org.Z) ) 
		return 0;
	//Minimize FPU usage, allocate needed stuff only
	FLOAT ActorX = Actor->Location.X - Org.X;
	FLOAT ActorY = Actor->Location.Y - Org.Y;
	FLOAT AdjustedActorY = ActorX * CollisionCacus->PrecomputedCoords.YAxis.X + ActorY * CollisionCacus->PrecomputedCoords.YAxis.Y;
	FLOAT TotalSide = Extent.X + Actor->CollisionRadius;
	//Check relative Y bounds, biggest reject chance (pawns are usually taller than wider)
	FLOAT XDeltaSq = Square(TotalSide) - Square(AdjustedActorY);
	if ( XDeltaSq < 0 )
		return 0;
	FLOAT AdjustedActorX = ActorX * CollisionCacus->PrecomputedCoords.XAxis.X + ActorY * CollisionCacus->PrecomputedCoords.XAxis.Y;
	//Cheap, non-precise high relative X bound check
	//Most horizontal traces are short movements, this will clear out a lot of cases
	if ( AdjustedActorX > Dist + TotalSide )
		return 0;
	//Refactored it looks like: (AdjX^2 + AdjY^ <= TS^2), this is a 'start inside actor' check
	if ( Square(AdjustedActorX) <= XDeltaSq )
	{
		Primitive = Cast<AXC_PrimitiveActor>(Actor->HitActor);
		if ( AdjustedActorX <= KINDA_SMALL_NUMBER )
		{
			if ( Primitive && Primitive->bBackTrace )
			{
				*HitLocation = Org;
				return 1;
			}
			return 0; //Tracing away from actor
		}
		*HitLocation = Org;
		*HitNormal = _UnsafeNormal2D( -ActorX, -ActorY);
		return 1;
	}
	//This is a 'actor behind trace' check
	if ( *(INT*)&AdjustedActorX < 0 ) //Is Negative check, avoid FPU stack load (do i get less cycles or what?)
		return 0;
	//Real X bound check
	//Move AdjX behind by corresponding cylinder extent
	FLOAT TargetX = AdjustedActorX - appSqrt( XDeltaSq);
	if ( TargetX > Dist )
		return 0;
	Primitive = Cast<AXC_PrimitiveActor>(Actor->HitActor); //Do this at the very end
	*HitLocation = Org + Dir * TargetX;
	*HitNormal = _UnsafeNormal2D( HitLocation->X - Actor->Location.X, HitLocation->Y - Actor->Location.Y);
	return 1;
	unguardf_debug( (Actor->GetName()) );
}

UBOOL FPrecomputedRay::HitsCylActor( AActor* Actor, FVector4* HitLocation, FVector4* HitNormal, AXC_PrimitiveActor* & Primitive)
{
	guard_debug(FPrecomputedRay::HitsCylActor);
	//Screw quick (X,Y,Z) rejects, I made the octree for this reason!!!
	FVector AdjustedActor = TransformPointByXY( CollisionCacus->PrecomputedCoords, Actor->Location ); //Gotta optimize that third coordinate
	//Check relative Y bounds, biggest reject chance
	FLOAT TotalSide = Extent.X + Actor->CollisionRadius;
	FLOAT XDeltaSq = Square(TotalSide) - Square(AdjustedActor.Y);
	if ( XDeltaSq < 0 )
		return 0;
	
	FLOAT ZDif = End.Z-Org.Z;
	FLOAT TotalHeight = Extent.Z + Actor->CollisionHeight;

	//Branchless Z bounds, 0 is low, 1 is high | One of them is always 0.f (before extent transformation)
	FLOAT ZBounds[2] = { -TotalHeight, TotalHeight};
	DWORD bUp = (ZDif >= 0); //See if ZDif transformation should go in high or low slot
	ZBounds[bUp] += ZDif;
	
	//Actor out of trace's Z bounds
	if ( AdjustedActor.Z < ZBounds[0] || AdjustedActor.Z > ZBounds[1] )
		return 0;
//	bUp = bUp << 31; //Move to sign bit
	
	//Trace starts inside actor?
	//Refactored it looks like: (AdjX^2 + AdjY^ <= TS^2) AND (within Z slabs)

	if ( Square(AdjustedActor.X) <= XDeltaSq ) //Contained in the infinite XY cylinder
	{
		if ( Abs(AdjustedActor.Z) <= TotalHeight ) //Within Z slab (in Cylinder)
		{
			Primitive = Cast<AXC_PrimitiveActor>(Actor->HitActor);
			if ( AdjustedActor.X <= KINDA_SMALL_NUMBER )
			{
				if ( Primitive && Primitive->bBackTrace )
				{
					*HitLocation = Org;
					return 1;
				}
				return 0; //Tracing away from actor
			}
			*HitLocation = Org;
			if ( Abs(AdjustedActor.Z * 0.98f) > TotalHeight ) //Up or down the cylinder
				*HitNormal = FVector(0,0,-fast_sign_nozero(AdjustedActor.Z));
			else //On side or inside
				*HitNormal = _UnsafeNormal2D( Org.X - Actor->Location.X, Org.Y - Actor->Location.Y );
			return 1;
		}
		if ( (*(INT*)&Dir.Z ^ *(INT*)&AdjustedActor.Z) < 0 ) //Not touching cylinder, negate if trace goes opposite sides
			return 0;
	}
	else //When start occurs outside of XY cylinder bounds (perform length checks)
	{
		//This is a 'actor behind trace' check
		if ( *(INT*)&AdjustedActor.X < 0 )
			return 0;
		//Lazy X bound check
		FLOAT XYDist = LengthUsingDir2D( &Org, &End, &CollisionCacus->PrecomputedCoords.XAxis);
		if ( AdjustedActor.X > XYDist + TotalSide ) 
			return 0;
		//Real X bound check
		FLOAT TargetX = AdjustedActor.X - appSqrt( XDeltaSq); //Move AdjX behind by corresponding cylinder extent
		if ( TargetX > XYDist )
			return 0;

		if ( Square(AdjustedActor.X) >= XDeltaSq ) //We can move forward and position ourselves next to the cylinder (side by side)
		{
			FLOAT Delta = _Reciprocal(Dir.Dot2( CollisionCacus->PrecomputedCoords.XAxis)) * TargetX; //1/HSize(Dir) * TargetX
			FLOAT ZEnd = Dir.Z * Delta;
			//Positioned point within Z slab, side hit
			if ( Square(ZEnd - AdjustedActor.Z) < Square(TotalHeight) )
			{
				Primitive = Cast<AXC_PrimitiveActor>(Actor->HitActor);
				HitLocation->X = Org.X + Dir.X * Delta;
				HitLocation->Y = Org.Y + Dir.Y * Delta;
				HitLocation->Z = Org.Z + ZEnd;
				*HitNormal = _UnsafeNormal2D( HitLocation->X - Actor->Location.X, HitLocation->Y - Actor->Location.Y);
				return 1;
			}
			if ( Square(ZEnd) > Square(AdjustedActor.Z) ) //We passed the actor already!
				return 0; //Appears to be an effective 'different sign' check
		}
	}

	
	FVector VStart(0,0,0);

//	if ( (AdjustedActor.Z) * ZDif <= 0 ) 
//		return 0; //Different Z directions, never gonna hit
		
	FLOAT PlaneZ = AdjustedActor.Z - fast_sign_nozero(AdjustedActor.Z) * TotalHeight;
//	VStart = Org + (End-Org) * (PlaneZ * _Reciprocal(ZDif));
	VStart = Org + (Dir) * (PlaneZ * Dist * _Reciprocal(ZDif));
	if ( (VStart - Actor->Location ).SizeSquared2D() > TotalSide*TotalSide )
		return 0;
	*HitLocation = VStart;
	*HitNormal = FVector(0,0,-fast_sign_nozero(ZDif));
	return 1;
	unguardf_debug( (Actor->GetName()) );
}

INT FPrecomputedRay::IntersectsBox( const FBox& Box)
{
	// Quick rejects, optimized for movement code.
	if( Box.Min.X > ExtendedBox.Max.X || ExtendedBox.Min.X > Box.Max.X
	||	Box.Min.Y > ExtendedBox.Max.Y || ExtendedBox.Min.Y > Box.Max.Y
	||	Box.Min.Z > ExtendedBox.Max.Z || ExtendedBox.Min.Z > Box.Max.Z )
	{
		return 0;
	}

//	FBox EBox( Box.Min-*Extent, Box.Max+*Extent);
	FBox EBox( Box.Min, Box.Max);
	if ( !Extent.IsZero() )
	{
		EBox.Min -= Extent;
		EBox.Max += Extent;
	}
	if ( FitsBox( Org, EBox) ) //This algorithm fails when traces start inside the box, let's cover this case
		return 1;

	//Back face culling
	//Checking on both inverses prevents early 'Dir' allocation in FPU registers
	FVector T;
	T.X = ((GEZero(Inv.X) ? EBox.Min.X : EBox.Max.X) - Org.X) * Inv.X;
	T.Y = ((GEZero(Inv.Y) ? EBox.Min.Y : EBox.Max.Y) - Org.Y) * Inv.Y;
	T.Z = ((GEZero(Inv.Z) ? EBox.Min.Z : EBox.Max.Z) - Org.Z) * Inv.Z;

	if ( (T.X >= T.Y) && (T.X >= T.Z) ) //X is maximum
	{
		if ( GEZero(T.X) /*&& (T.X <= Dist) */)
		{
			FLOAT ptY = Org.Y + Dir.Y * T.X;
			FLOAT ptZ = Org.Z + Dir.Z * T.X;
			return ptY >= EBox.Min.Y && ptY <= EBox.Max.Y
				&& ptZ >= EBox.Min.Z && ptZ <= EBox.Max.Z;
		}
	}
	else if ( T.Y >= T.Z ) //Y is maximum
	{
		if ( GEZero(T.Y) /*&& (T.Y <= Dist)*/ )
		{
			FLOAT ptX = Org.X + Dir.X * T.Y;
			FLOAT ptZ = Org.Z + Dir.Z * T.Y;
			return ptX >= EBox.Min.X && ptX <= EBox.Max.X
				&& ptZ >= EBox.Min.Z && ptZ <= EBox.Max.Z;
		}
	}
	else //Z is maximum
	{
		if ( GEZero(T.Z) /*&& (T.Z <= Dist)*/ )
		{
			FLOAT ptX = Org.X + Dir.X * T.Z;
			FLOAT ptY = Org.Y + Dir.Y * T.Z;
			return ptX >= EBox.Min.X && ptX <= EBox.Max.X
				&& ptY >= EBox.Min.Y && ptY <= EBox.Max.Y;
		}
	}
	return 0;
}


//Aligned SSE box generation
inline void TwoPointBox( FBox& Box, const FVector4& A, const FVector4& B, const FVector4& Extent)
{
#if ASM
		__asm
		{
			mov		eax,[A] //Was Other
			mov		ecx,[B] //Was this
			mov		edx,[Box]
			mov		ebx,[Extent]

			movaps	xmm0,[eax]	//A
			movaps	xmm1,[ecx]	//B
			movaps	xmm3,[ebx]	//EXTENT

			movaps	xmm2,xmm1	//B
			minps	xmm2,xmm0	//Min(A,B)
			subps	xmm2,xmm3	//Min(A,B) - Extent
			movaps	[edx],xmm2	//Store | use align due to previous aligned data

			movaps	xmm2,xmm1	//B
			maxps	xmm2,xmm0	//Max(A,B)
			addps	xmm2,xmm3	//Max(A,B) + Extent
			movups	[edx+12],xmm2	//Store (extra 4 bytes overwritten!!!
		}
#else
	if ( A.X > B.X )
	{
		Box.Min.X = B.X - Extent.X;
		Box.Max.X = A.X + Extent.X;
	}
	else
	{
		Box.Min.X = A.X - Extent.X;
		Box.Max.X = B.X + Extent.X;
	}
	if ( A.Y > B.Y )
	{
		Box.Min.Y = B.Y - Extent.Y;
		Box.Max.Y = A.Y + Extent.Y;
	}
	else
	{
		Box.Min.Y = A.Y - Extent.Y;
		Box.Max.Y = B.Y + Extent.Y;
	}
	if ( A.Z > B.Z )
	{
		Box.Min.Z = B.Z - Extent.Z;
		Box.Max.Z = A.Z + Extent.Z;
	}
	else
	{
		Box.Min.Z = A.Z - Extent.Z;
		Box.Max.Z = B.Z + Extent.Z;
	}
#endif
	Box.IsValid = 1;
}

//============================================================================//
//============================================================================//
//===                          FNodeGrid3 methods                          ===//
//============================================================================//
//============================================================================//


void FNodeGrid3::SetBaseGrid( FCollisionOctree* Base)
{
	GridSize = 1;
	Grid[0] = Base;

	BoxSize = Base->BoundingBox.Max - Base->BoundingBox.Min;
	Divider = FVector( _Reciprocal(BoxSize.X), _Reciprocal(BoxSize.Y), _Reciprocal(BoxSize.Z) );
	Substract = Base->BoundingBox.Min;
}

void FNodeGrid3::SetGrid( FNodeGrid3* Upper)
{
	GridSize = Upper->GridSize << 1;
	Substract = Upper->Substract;
	BoxSize = Upper->BoxSize * 0.5f;
	Divider = Upper->Divider * 2.f;
//	debugf( NAME_Log, TEXT("TEST: Box=%f, Divider=%f, GridSize=%f, Sub=%f"), BoxSize.X, Divider.X, (FLOAT)GridSize, Substract.X);
}

void FNodeGrid3::Register( FCollisionOctree* In)
{
	INT X = In->XPos;
	INT Y = In->YPos;
	INT Z = In->ZPos;
	INT Point = X + Y * GridSize + Z * GridSize * GridSize;

	if ( Grid[Point] )
	{
		debugf( NAME_Log, TEXT("CRITICAL NODE REDUNDANCY ERROR:") );
		debugf( NAME_Log, TEXT("===> Min: (%i,%i,%i)/(%i,%i,%i)"), appRound(Grid[Point]->BoundingBox.Min.X),appRound(Grid[Point]->BoundingBox.Min.Y),appRound(Grid[Point]->BoundingBox.Min.Z), appRound(In->BoundingBox.Min.X),appRound(In->BoundingBox.Min.Y),appRound(In->BoundingBox.Min.Z));
		debugf( NAME_Log, TEXT("===> Max: (%i,%i,%i)/(%i,%i,%i)"), appRound(Grid[Point]->BoundingBox.Max.X),appRound(Grid[Point]->BoundingBox.Max.Y),appRound(Grid[Point]->BoundingBox.Max.Z), appRound(In->BoundingBox.Max.X),appRound(In->BoundingBox.Max.Y),appRound(In->BoundingBox.Max.Z));
		debugf( NAME_Log, TEXT("Register: Depth=%i, (%i,%i,%i) to %i"), In->CurDepth, X, Y, Z, In->MIndex);
	}
	Grid[Point] = In;
	In->MIndex = In->Master->GridData->Position[In->CurDepth] + Point;
}


void FNodeGrid3::Remove( FCollisionOctree* In)
{
	if ( !Grid ) //Something destroyed this first
		return;
	INT X = In->XPos;		INT Y = In->YPos;		INT Z = In->ZPos;
	INT Point = X + Y * GridSize + Z * GridSize * GridSize;
	Grid[Point] = NULL;
}

INT FNodeGrid3::GetIdx( const FVector& Pos)
{
	INT Point = Clamp( appFloor((Pos.X-Substract.X) * Divider.X), 0, GridSize-1);
	Point += Clamp( appFloor((Pos.Y-Substract.Y) * Divider.Y), 0, GridSize-1) * GridSize;
	Point += Clamp( appFloor((Pos.Y-Substract.Z) * Divider.Z), 0, GridSize-1) * GridSize * GridSize;
	return Point;
}

void FNodeGrid3::GetXYZ( const FVector& Pos, INT* XYZ)
{
	INT CLMax = GridSize-1;
	XYZ[0] = Clamp( appFloor((Pos.X-Substract.X) * Divider.X), 0, CLMax);
	XYZ[1] = Clamp( appFloor((Pos.Y-Substract.Y) * Divider.Y), 0, CLMax);
	XYZ[2] = Clamp( appFloor((Pos.Z-Substract.Z) * Divider.Y), 0, CLMax);
}

#endif
